package ru.dbotthepony.mc.otm;

import com.mojang.blaze3d.systems.RenderSystem;
import net.minecraft.client.Minecraft;
import net.minecraft.client.gui.screens.MenuScreens;
import net.minecraft.client.player.LocalPlayer;
import net.minecraft.resources.ResourceLocation;
import net.minecraft.server.level.ServerPlayer;
import net.minecraft.world.entity.Entity;
import net.minecraft.world.entity.LivingEntity;
import net.minecraft.world.entity.player.Player;
import net.minecraft.world.inventory.MenuType;
import net.minecraft.world.item.BlockItem;
import net.minecraft.world.item.CreativeModeTab;
import net.minecraft.world.item.Item;
import net.minecraft.world.level.block.Block;
import net.minecraft.world.level.block.entity.BlockEntityType;
import net.minecraftforge.client.event.RenderGameOverlayEvent;
import net.minecraftforge.client.gui.ForgeIngameGui;
import net.minecraftforge.common.MinecraftForge;
import net.minecraftforge.common.util.INBTSerializable;
import net.minecraftforge.common.util.LazyOptional;
import net.minecraftforge.event.AttachCapabilitiesEvent;
import net.minecraftforge.event.RegistryEvent;
import net.minecraftforge.event.TickEvent;
import net.minecraftforge.event.entity.living.LivingDeathEvent;
import net.minecraftforge.event.entity.living.LivingEvent;
import net.minecraftforge.event.entity.player.PlayerEvent;
import net.minecraftforge.eventbus.api.SubscribeEvent;
import net.minecraftforge.fml.InterModComms;
import net.minecraftforge.fml.common.Mod;
import net.minecraftforge.fml.event.lifecycle.FMLCommonSetupEvent;
import net.minecraftforge.fml.event.lifecycle.InterModEnqueueEvent;
import net.minecraftforge.fml.event.lifecycle.InterModProcessEvent;
import net.minecraftforge.fmlserverevents.FMLServerStartingEvent;
import net.minecraftforge.fml.javafmlmod.FMLJavaModLoadingContext;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import ru.dbotthepony.mc.otm.block.BlockAndroidStation;
import ru.dbotthepony.mc.otm.block.entity.BlockEntityAndroidStation;
import ru.dbotthepony.mc.otm.capability.AndroidCapability;
import ru.dbotthepony.mc.otm.capability.AndroidCapabilityPlayer;
import ru.dbotthepony.mc.otm.capability.IAndroidCapability;
import ru.dbotthepony.mc.otm.capability.MatteryCapability;
import ru.dbotthepony.mc.otm.menu.AndroidStationMenu;
import ru.dbotthepony.mc.otm.network.MatteryNetwork;
import ru.dbotthepony.mc.otm.screen.AndroidStationScreen;

import java.math.BigDecimal;
import java.util.Optional;
import java.util.stream.Collectors;

// The value here should match an entry in the META-INF/mods.toml file
@Mod(OverdriveThatMatters.MOD_ID)
public class OverdriveThatMatters {
	// Directly reference a log4j logger.
	public static final String MOD_ID = "overdrive_that_matters";
	public static final Logger LOGGER = LogManager.getLogger();

	public static final ResourceLocation PLAYER_GUI_LOCATION = new ResourceLocation(OverdriveThatMatters.MOD_ID, "textures/gui/player_gui.png");

	public static Block ANDROID_STATION = new BlockAndroidStation();
	public static BlockEntityType.Builder<BlockEntityAndroidStation> ANDROID_STATION_BUILDER = BlockEntityType.Builder.of(BlockEntityAndroidStation::new, ANDROID_STATION);
	public static BlockEntityType<BlockEntityAndroidStation> ANDROID_STATION_FACTORY = null;
	public static Item ANDROID_STATION_ITEM = new BlockItem(ANDROID_STATION, new Item.Properties().stacksTo(64).tab(CreativeModeTab.TAB_MISC)).setRegistryName(BlockAndroidStation.REGISTRY_NAME);
	public static MenuType<AndroidStationMenu> ANDROID_STATION_MENU_TYPE = new MenuType<>(AndroidStationMenu::new);

	static {
		ANDROID_STATION_MENU_TYPE.setRegistryName(BlockAndroidStation.REGISTRY_NAME);
	}

	// public static final Block STONE = register("stone", new Block(BlockBehaviour.Properties.of(Material.STONE, MaterialColor.STONE).requiresCorrectToolForDrops().strength(1.5F, 6.0F)));

	public OverdriveThatMatters() {
		// Register the setup method for modloading
		FMLJavaModLoadingContext.get().getModEventBus().addListener(this::setup);
		// Register the enqueueIMC method for modloading
		FMLJavaModLoadingContext.get().getModEventBus().addListener(this::enqueueIMC);
		// Register the processIMC method for modloading
		FMLJavaModLoadingContext.get().getModEventBus().addListener(this::processIMC);

		// Register ourselves for server and other game events we are interested in
		MinecraftForge.EVENT_BUS.register(this);
	}

	private void setup(final FMLCommonSetupEvent event) {
		// some preinit code
		MatteryCapability.register();
		MatteryNetwork.register();
	}

	private void enqueueIMC(final InterModEnqueueEvent event) {
		// some example code to dispatch IMC to another mod
		InterModComms.sendTo("examplemod", "helloworld", () -> { LOGGER.info("Hello world from the MDK"); return "Hello world";});
	}

	private void processIMC(final InterModProcessEvent event) {
		// some example code to receive and process InterModComms from other mods
		LOGGER.info("Got IMC {}", event.getIMCStream().
				map(m->m.messageSupplier().get()).
				collect(Collectors.toList()));
	}

	// You can use SubscribeEvent and let the Event Bus discover methods to call
	@SubscribeEvent
	public void onServerStarting(FMLServerStartingEvent event) {
		// do something when the server starts
	}

	private final ResourceLocation android_cap_location = new ResourceLocation(OverdriveThatMatters.MOD_ID, "android_capability");

	@SubscribeEvent
	public void onAttachCapabilityEvent(AttachCapabilitiesEvent<Entity> event) {
		if (!(event.getObject() instanceof Player)) {
			return;
		}

		event.addCapability(android_cap_location, new AndroidCapabilityPlayer((Player) event.getObject()));
	}

	@SubscribeEvent
	public void onLivingTick(LivingEvent.LivingUpdateEvent event) {
		if (event.getEntity() instanceof ServerPlayer ply) {
			ply.getCapability(MatteryCapability.ANDROID).ifPresent(IAndroidCapability::tick);
		} else if (event.getEntity() instanceof LocalPlayer ply) {
			ply.getCapability(MatteryCapability.ANDROID).ifPresent(IAndroidCapability::tickClient);
		}
	}

	@SubscribeEvent
	public void onPlayerCloneEvent(PlayerEvent.Clone event) {
		event.getPlayer().getCapability(MatteryCapability.ANDROID).ifPresent((cap) -> {
			LazyOptional<IAndroidCapability> resolver = event.getOriginal().getCapability(MatteryCapability.ANDROID);

			if (!resolver.isPresent() || resolver.resolve().isEmpty()) {
				event.getOriginal().reviveCaps();
				resolver = event.getOriginal().getCapability(MatteryCapability.ANDROID);
				LOGGER.info("revived cap {}", resolver.resolve());
			}

			if (!resolver.isPresent() || resolver.resolve().isEmpty()) {
				event.getOriginal().invalidateCaps();
				return;
			}

			cap.deserializeNBT(resolver.resolve().get().serializeNBT());
			event.getOriginal().invalidateCaps();
		});
	}

	@SubscribeEvent
	public void onLayerRenderEvent(RenderGameOverlayEvent.PreLayer event) {
		if (event.getOverlay() != ForgeIngameGui.FOOD_LEVEL_ELEMENT)
			return;

		Minecraft mc = Minecraft.getInstance();
		LocalPlayer ply = mc.player;

		if (ply == null)
			return;

		if (
			ply.getVehicle() instanceof LivingEntity ||
			mc.options.hideGui ||
			!mc.gameMode.canHurtPlayer() ||
			!(mc.getCameraEntity() instanceof Player)
		)
			return;

		Optional<IAndroidCapability> lazy = ply.getCapability(MatteryCapability.ANDROID).resolve();

		if (lazy.isPresent() && mc.gui instanceof ForgeIngameGui gui) {
			event.setCanceled(true);
			IAndroidCapability android = lazy.get();

			float level;

			if (android.getMaxBatteryLevel().compareTo(BigDecimal.ZERO) == 0) {
				level = 0f;
			} else {
				level = android.getBatteryLevel().divide(android.getMaxBatteryLevel(), MatteryCapability.ROUND_RULES).floatValue();

				if (level >= 0.98f)
					level = 1f;
			}

			gui.setupOverlayRenderState(true, false);

			RenderSystem.setShaderColor(1.0F, 1.0F, 1.0F, 1.0F);
			RenderSystem.setShaderTexture(0, PLAYER_GUI_LOCATION);

			int width = event.getWindow().getGuiScaledWidth();
			int height = event.getWindow().getGuiScaledHeight();

			int left = width / 2 + 10;
			int top = height - gui.right_height;

			gui.right_height += 10;

			// Stack, x, y, blitOffset?, (float) image_x, (float) image_y, rect_x, rect_y, total_image_width, total_image_height
			// Stack, x, y, image_x, image_y, rect_x, rect_y
			gui.blit(event.getMatrixStack(), left, top, 0, 0, 80, 9);
			int shift_left = (int) Math.ceil(level * 80f - 0.5f);
			gui.blit(event.getMatrixStack(), left + 80 - shift_left, top, 0, 9, shift_left, 9);
		}
	}

	// You can use EventBusSubscriber to automatically subscribe events on the contained class (this is subscribing to the MOD
	// Event bus for receiving Registry Events)
	@Mod.EventBusSubscriber(bus=Mod.EventBusSubscriber.Bus.MOD)
	public static class RegistryEvents {
		@SubscribeEvent
		public static void onBlocksRegistry(final RegistryEvent.Register<Block> blockRegistryEvent) {
			blockRegistryEvent.getRegistry().register(ANDROID_STATION);
		}

		@SubscribeEvent
		public static void onItemRegistry(final RegistryEvent.Register<Item> blockRegistryEvent) {
			blockRegistryEvent.getRegistry().register(ANDROID_STATION_ITEM);
		}

		@SubscribeEvent
		public static void onBlockEntityRegistry(final RegistryEvent.Register<BlockEntityType<?>> blockRegistryEvent) {
			ANDROID_STATION_FACTORY = ANDROID_STATION_BUILDER.build(null);
			ANDROID_STATION_FACTORY.setRegistryName(BlockAndroidStation.REGISTRY_NAME);
			blockRegistryEvent.getRegistry().register(ANDROID_STATION_FACTORY);
		}

		@SubscribeEvent
		public static void onMenuTypeRegistry(final RegistryEvent.Register<MenuType<?>> blockRegistryEvent) {
			blockRegistryEvent.getRegistry().register(ANDROID_STATION_MENU_TYPE);
			MenuScreens.register(ANDROID_STATION_MENU_TYPE, AndroidStationScreen::new);
		}
	}
}
